@@ -9,6 +9,7 @@ use dbus::{
     Message,
 };
 use dbus_tree::{MTSync, MethodErr, MethodInfo, MethodResult, PropInfo, Tree};
+use futures::executor::block_on;
 
 use crate::{
     dbus_api::{
@@ -17,7 +18,7 @@ use crate::{
         types::{DbusErrorEnum, TData, OK_STRING},
         util::{engine_to_dbus_err_tuple, get_next_arg},
     },
-    engine::{BlockDevTier, Engine, EngineAction, Name, Pool, PoolUuid},
+    engine::{BlockDevTier, Engine, EngineAction, LockKey, Name, Pool, PoolUuid},
 };
 
 pub enum BlockDevOp {
@@ -51,10 +52,9 @@ where
         Pool
     );
 
-    let mutex_lock = dbus_context.engine.blocking_lock();
-    let (pool_name, pool) = mutex_lock
-        .get_pool(pool_uuid)
+    let guard = block_on(dbus_context.engine.get_pool(LockKey::Uuid(pool_uuid)))
         .ok_or_else(|| format!("no pool corresponding to uuid {}", &pool_uuid))?;
+    let (pool_name, _, pool) = guard.as_tuple();
 
     closure((pool_name, pool_uuid, pool))
 }
@@ -93,8 +93,8 @@ where
         return_message
     );
 
-    let mut mutex_lock = dbus_context.engine.blocking_lock();
-    let (pool_name, pool) = get_mut_pool!(mutex_lock; pool_uuid; default_return; return_message);
+    let guard = get_mut_pool!(dbus_context.engine; pool_uuid; default_return; return_message);
+    let (pool_name, _, pool) = guard.as_tuple();
 
     let blockdevs = devs.map(Path::new).collect::<Vec<&Path>>();
 
@@ -108,20 +108,34 @@ where
             dbus_context.push_pool_cache_change(pool_path.get_name(), true);
             res
         }
-        BlockDevOp::AddCache => handle_action!(
-            pool.add_blockdevs(pool_uuid, &*pool_name, &blockdevs, BlockDevTier::Cache),
-            dbus_context,
-            pool_path.get_name()
-        ),
-        BlockDevOp::AddData => {
-            let res = handle_action!(
-                pool.add_blockdevs(pool_uuid, &*pool_name, &blockdevs, BlockDevTier::Data),
+        BlockDevOp::AddCache => {
+            handle_action!(
+                pool.add_blockdevs(pool_uuid, &*pool_name, &blockdevs, BlockDevTier::Cache,)
+                    .map(|(act, _)| { act }),
                 dbus_context,
                 pool_path.get_name()
-            );
-            dbus_context
-                .push_pool_size_change(pool_path.get_name(), pool.total_physical_size().bytes());
-            res
+            )
+        }
+        BlockDevOp::AddData => {
+            handle_action!(
+                pool.add_blockdevs(pool_uuid, &*pool_name, &blockdevs, BlockDevTier::Data,)
+                    .map(|(act, diff)| {
+                        dbus_context.push_pool_size_change(
+                            pool_path.get_name(),
+                            pool.total_physical_size().bytes(),
+                        );
+                        if let Some(used) = diff.as_ref().and_then(|d| d.usage) {
+                            dbus_context.push_pool_used_change(pool_path.get_name(), used);
+                        }
+                        if let Some(alloc) = diff.and_then(|d| d.thin_pool.allocated_size) {
+                            dbus_context
+                                .push_pool_allocated_size_change(pool_path.get_name(), alloc);
+                        }
+                        act
+                    }),
+                dbus_context,
+                pool_path.get_name()
+            )
         }
     };
     let msg = match result.map(|bds| bds.changed()) {
@@ -250,7 +264,7 @@ pub fn pool_used_size<E>(pool: &E::Pool) -> (bool, String)
 where
     E: 'static + Engine,
 {
-    prop_conv::pool_used_to_prop(pool.total_physical_used().map(|u| u.bytes()).ok())
+    prop_conv::pool_used_to_prop(pool.total_physical_used().map(|u| u.bytes()))
 }
 
 /// Generate a D-Bus representation of the total size of the pool in bytes.
