@@ -2,9 +2,19 @@
 // License, v. 2.0. If a copy of the MPL was not distributed with this
 // file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
-use std::{collections::HashMap, path::Path};
+use std::{
+    collections::{HashMap, HashSet},
+    path::Path,
+    sync::Arc,
+};
 
+use async_trait::async_trait;
+use futures::{executor::block_on, future::join_all};
 use serde_json::Value;
+use tokio::{
+    sync::{OwnedRwLockReadGuard, OwnedRwLockWriteGuard, RwLock},
+    task::{spawn_blocking, JoinHandle},
+};
 
 use devicemapper::DmNameBuf;
 
@@ -18,31 +28,36 @@ use crate::{
             liminal::{find_all, LiminalDevices},
             pool::StratPool,
         },
-        structures::Table,
+        structures::{
+            AllLockReadGuard, AllLockWriteGuard, AllOrSomeLock, ExclusiveGuard, Lockable,
+            SharedGuard, SomeLockReadGuard, SomeLockWriteGuard, Table,
+        },
         types::{
-            CreateAction, DeleteAction, DevUuid, EncryptionInfo, FilesystemUuid, LockedPoolInfo,
-            RenameAction, ReportType, SetUnlockAction, StratFilesystemDiff, ThinPoolDiff,
-            UdevEngineEvent, UnlockMethod,
+            CreateAction, DeleteAction, DevUuid, EncryptionInfo, FilesystemUuid, LockKey,
+            LockedPoolInfo, RenameAction, ReportType, SetUnlockAction, StratFilesystemDiff,
+            StratPoolDiff, UdevEngineEvent, UnlockMethod,
         },
         Engine, Name, PoolUuid, Report,
     },
     stratis::{StratisError, StratisResult},
 };
 
+type EventNumbers = HashMap<PoolUuid, HashMap<DmNameBuf, u32>>;
+
 #[derive(Debug)]
 pub struct StratEngine {
-    pools: Table<PoolUuid, StratPool>,
+    pools: AllOrSomeLock<PoolUuid, StratPool>,
 
     // Maps pool UUIDs to information about sets of devices that are
     // associated with that UUID but have not been converted into a pool.
-    liminal_devices: LiminalDevices,
+    liminal_devices: Lockable<Arc<RwLock<LiminalDevices>>>,
 
     // Maps name of DM devices we are watching to the most recent event number
     // we've handled for each
-    watched_dev_last_event_nrs: HashMap<PoolUuid, HashMap<DmNameBuf, u32>>,
+    watched_dev_last_event_nrs: Lockable<Arc<RwLock<EventNumbers>>>,
 
     // Handler for key operations
-    key_handler: StratKeyActions,
+    key_handler: Lockable<Arc<RwLock<StratKeyActions>>>,
 
     // TODO: Remove this code when Clevis supports reading keys from the
     // kernel keyring.
@@ -52,6 +67,8 @@ pub struct StratEngine {
     key_fs: MemoryFilesystem,
 }
 
+type PoolJoinHandles = Vec<JoinHandle<StratisResult<Option<(PoolUuid, StratPoolDiff)>>>>;
+
 impl StratEngine {
     /// Setup a StratEngine.
     /// 1. Verify the existence of Stratis /dev directory.
@@ -72,20 +89,153 @@ impl StratEngine {
         }
 
         Ok(StratEngine {
-            pools,
-            liminal_devices,
-            watched_dev_last_event_nrs: HashMap::new(),
-            key_handler: StratKeyActions,
+            pools: AllOrSomeLock::new(pools),
+            liminal_devices: Lockable::new_shared(liminal_devices),
+            watched_dev_last_event_nrs: Lockable::new_shared(HashMap::new()),
+            key_handler: Lockable::new_shared(StratKeyActions),
             key_fs: MemoryFilesystem::new()?,
         })
     }
 
+    fn spawn_pool_check_handling(
+        joins: &mut PoolJoinHandles,
+        guard: SomeLockWriteGuard<PoolUuid, StratPool>,
+    ) {
+        joins.push(spawn_blocking(move || {
+            let (name, uuid, pool) = guard.as_tuple();
+            let diff = pool.event_on(uuid, &name)?;
+            Ok(if diff.is_changed() {
+                Some((uuid, diff))
+            } else {
+                None
+            })
+        }));
+    }
+
+    fn spawn_fs_check_handling(
+        joins: &mut Vec<JoinHandle<StratisResult<HashMap<FilesystemUuid, StratFilesystemDiff>>>>,
+        guard: SomeLockWriteGuard<PoolUuid, StratPool>,
+    ) {
+        joins.push(spawn_blocking(move || {
+            let (name, uuid, pool) = guard.as_tuple();
+            pool.fs_event_on(uuid, &name)
+        }));
+    }
+
+    async fn join_all_pool_checks(handles: PoolJoinHandles) -> HashMap<PoolUuid, StratPoolDiff> {
+        join_all(handles)
+            .await
+            .into_iter()
+            .filter_map(|res| match res {
+                Ok(Ok(opt)) => opt,
+                Ok(Err(e)) => {
+                    warn!("Pool checks failed with error: {}", e);
+                    None
+                }
+                Err(e) => {
+                    warn!(
+                        "Failed to get status for thread handling pool checks: {}",
+                        e
+                    );
+                    None
+                }
+            })
+            .collect::<HashMap<_, _>>()
+    }
+
+    async fn join_all_fs_checks(
+        handles: Vec<JoinHandle<StratisResult<HashMap<FilesystemUuid, StratFilesystemDiff>>>>,
+    ) -> HashMap<FilesystemUuid, StratFilesystemDiff> {
+        join_all(handles)
+            .await
+            .into_iter()
+            .fold(HashMap::default(), |mut acc, next| match next {
+                Ok(Ok(hm)) => {
+                    acc.extend(hm);
+                    acc
+                }
+                Ok(Err(e)) => {
+                    warn!("Filesystem checks failed with error: {}", e);
+                    acc
+                }
+                Err(e) => {
+                    warn!(
+                        "Failed to get status for thread handling filesystem checks: {}",
+                        e
+                    );
+                    acc
+                }
+            })
+    }
+
+    /// The implementation for pool_evented when caused by a devicemapper event.
+    async fn pool_evented_dm(&self, pools: &HashSet<PoolUuid>) -> HashMap<PoolUuid, StratPoolDiff> {
+        let mut joins = Vec::new();
+        for uuid in pools {
+            if let Some(guard) = self.pools.write(LockKey::Uuid(*uuid)).await {
+                Self::spawn_pool_check_handling(&mut joins, guard);
+            } else {
+                warn!(
+                    "Pool with UUID {} indicated an event but could not be found when attempting to handle the event; it may have been deleted during the time between detection and handling",
+                    uuid
+                );
+            }
+        }
+
+        Self::join_all_pool_checks(joins).await
+    }
+
+    /// The implementation for pool_evented when called by the timer thread.
+    async fn pool_evented_timer(&self) -> HashMap<PoolUuid, StratPoolDiff> {
+        let mut joins = Vec::new();
+        let guards: Vec<SomeLockWriteGuard<PoolUuid, StratPool>> =
+            self.pools.write_all().await.into();
+        for guard in guards {
+            Self::spawn_pool_check_handling(&mut joins, guard);
+        }
+
+        Self::join_all_pool_checks(joins).await
+    }
+
+    /// The implementation for fs_evented when caused by a devicemapper event.
+    async fn fs_evented_dm(
+        &self,
+        pools: &HashSet<PoolUuid>,
+    ) -> HashMap<FilesystemUuid, StratFilesystemDiff> {
+        let mut joins = Vec::new();
+        for uuid in pools {
+            if let Some(guard) = self.pools.write(LockKey::Uuid(*uuid)).await {
+                Self::spawn_fs_check_handling(&mut joins, guard);
+            } else {
+                warn!(
+                    "Pool with UUID {} indicated an event but could not be found when attempting to handle the event; it may have been deleted during the time between detection and handling",
+                    uuid
+                );
+            }
+        }
+
+        Self::join_all_fs_checks(joins).await
+    }
+
+    /// The implementation for fs_evented when called by the timer thread.
+    async fn fs_evented_timer(&self) -> HashMap<FilesystemUuid, StratFilesystemDiff> {
+        let mut joins = Vec::new();
+        let guards: Vec<SomeLockWriteGuard<PoolUuid, StratPool>> =
+            self.pools.write_all().await.into();
+        for guard in guards {
+            Self::spawn_fs_check_handling(&mut joins, guard);
+        }
+
+        Self::join_all_fs_checks(joins).await
+    }
+
     /// Recursively remove all devicemapper devices in all pools.
     /// Do not remove the dm-crypt devices that comprise the backstore.
     #[cfg(test)]
     pub fn teardown(self) -> StratisResult<()> {
         let mut untorndown_pools = Vec::new();
-        for (_, uuid, mut pool) in self.pools {
+        let mut write_all = block_on(self.pools.write_all());
+        for (_, uuid, pool) in write_all.iter_mut() {
             pool.teardown()
                 .unwrap_or_else(|_| untorndown_pools.push(uuid));
         }
@@ -107,7 +257,7 @@ impl<'a> Into<Value> for &'a StratEngine {
     fn into(self) -> Value {
         let json = json!({
             "pools": Value::Array(
-                self.pools.iter()
+                block_on(self.pools.read_all()).iter()
                     .map(|(name, uuid, pool)| {
                         let mut json = json!({
                             "uuid": Value::from(uuid.to_string()),
@@ -131,7 +281,7 @@ impl<'a> Into<Value> for &'a StratEngine {
         });
         if let (Value::Object(mut j), Value::Object(map)) = (
             json,
-            <&LiminalDevices as Into<Value>>::into(&self.liminal_devices),
+            <&LiminalDevices as Into<Value>>::into(&*self.liminal_devices.blocking_read()),
         ) {
             j.extend(map.into_iter());
             Value::Object(j)
@@ -148,64 +298,92 @@ impl Report for StratEngine {
 
     fn get_report(&self, report_type: ReportType) -> Value {
         match report_type {
-            ReportType::ErroredPoolDevices => (&self.liminal_devices).into(),
+            ReportType::ErroredPoolDevices => (&*self.liminal_devices.blocking_read()).into(),
         }
     }
 }
 
+#[async_trait]
 impl Engine for StratEngine {
     type Pool = StratPool;
     type KeyActions = StratKeyActions;
 
-    fn handle_event(&mut self, event: &UdevEngineEvent) -> Option<(Name, PoolUuid, &Self::Pool)> {
-        if let Some((pool_uuid, pool_name, pool)) =
-            self.liminal_devices.block_evaluate(&self.pools, event)
-        {
-            self.pools.insert(pool_name.clone(), pool_uuid, pool);
-            Some((
-                pool_name,
-                pool_uuid,
-                self.pools.get_by_uuid(pool_uuid).expect("just_inserted").1,
-            ))
-        } else {
-            None
+    async fn handle_events(
+        &self,
+        events: Vec<UdevEngineEvent>,
+    ) -> Vec<SomeLockReadGuard<PoolUuid, Self::Pool>> {
+        let mut ret_guards = Vec::new();
+        let uuids = {
+            let mut ld_guard = self.liminal_devices.write().await;
+            let mut pools_write_all = self.pools.write_all().await;
+            match spawn_blocking!({
+                events
+                    .into_iter()
+                    .filter_map(|event| {
+                        if let Some((uuid, name, pool)) =
+                            ld_guard.block_evaluate(&*pools_write_all, &event)
+                        {
+                            pools_write_all.insert(name, uuid, pool);
+                            Some(uuid)
+                        } else {
+                            None
+                        }
+                    })
+                    .collect::<Vec<_>>()
+            }) {
+                Ok(u) => u,
+                Err(e) => {
+                    warn!("Failed to handle udev events: {}", e);
+                    return ret_guards;
+                }
+            }
+        };
+        for uuid in uuids {
+            if let Some(guard) = self.pools.read(LockKey::Uuid(uuid)).await {
+                ret_guards.push(guard);
+            }
         }
+        ret_guards
     }
 
-    fn create_pool(
-        &mut self,
+    async fn create_pool(
+        &self,
         name: &str,
         blockdev_paths: &[&Path],
-        redundancy: Option<u16>,
         encryption_info: Option<&EncryptionInfo>,
     ) -> StratisResult<CreateAction<PoolUuid>> {
-        let redundancy = calculate_redundancy!(redundancy);
-
         validate_name(name)?;
+        let name = Name::new(name.to_owned());
 
         validate_paths(blockdev_paths)?;
 
-        match self.pools.get_by_name(name) {
-            Some((_, pool)) => create_pool_idempotent_or_err(pool, name, blockdev_paths),
-            None => {
-                if blockdev_paths.is_empty() {
-                    Err(StratisError::Msg(
-                        "At least one blockdev is required to create a pool.".to_string(),
-                    ))
-                } else {
-                    let (uuid, pool) =
-                        StratPool::initialize(name, blockdev_paths, redundancy, encryption_info)?;
+        let maybe_guard = self.pools.read(LockKey::Name(name.clone())).await;
+        if let Some(guard) = maybe_guard {
+            let (name, _, pool) = guard.as_tuple();
+            create_pool_idempotent_or_err(pool, &name, blockdev_paths)
+        } else if blockdev_paths.is_empty() {
+            Err(StratisError::Msg(
+                "At least one blockdev is required to create a pool.".to_string(),
+            ))
+        } else {
+            let cloned_name = name.clone();
+            let cloned_paths = blockdev_paths
+                .iter()
+                .map(|p| p.to_path_buf())
+                .collect::<Vec<_>>();
+            let cloned_enc_info = encryption_info.cloned();
 
-                    let name = Name::new(name.to_owned());
-                    self.pools.insert(name, uuid, pool);
-                    Ok(CreateAction::Created(uuid))
-                }
-            }
+            let (pool_uuid, _) = spawn_blocking!({
+                let borrowed_paths = cloned_paths.iter().map(|p| p.as_path()).collect::<Vec<_>>();
+                StratPool::initialize(&cloned_name, &borrowed_paths, cloned_enc_info.as_ref())
+            })??;
+
+            Ok(CreateAction::Created(pool_uuid))
         }
     }
 
-    fn destroy_pool(&mut self, uuid: PoolUuid) -> StratisResult<DeleteAction<PoolUuid>> {
-        if let Some((_, pool)) = self.pools.get_by_uuid(uuid) {
+    async fn destroy_pool(&self, uuid: PoolUuid) -> StratisResult<DeleteAction<PoolUuid>> {
+        if let Some(pool) = self.pools.read(LockKey::Uuid(uuid)).await {
             if pool.has_filesystems() {
                 return Err(StratisError::Msg("filesystems remaining on pool".into()));
             };
@@ -213,82 +391,87 @@ impl Engine for StratEngine {
             return Ok(DeleteAction::Identity);
         }
 
-        let (pool_name, mut pool) = self
-            .pools
+        let mut guard = self.pools.write_all().await;
+        let (pool_name, mut pool) = guard
             .remove_by_uuid(uuid)
             .expect("Must succeed since self.pools.get_by_uuid() returned a value");
 
-        if let Err(err) = pool.destroy() {
-            self.pools.insert(pool_name, uuid, pool);
+        let (res, pool) = spawn_blocking!((pool.destroy(), pool))?;
+        if let Err(err) = res {
+            guard.insert(pool_name, uuid, pool);
             Err(err)
         } else {
             Ok(DeleteAction::Deleted(uuid))
         }
     }
 
-    fn rename_pool(
-        &mut self,
+    async fn rename_pool(
+        &self,
         uuid: PoolUuid,
         new_name: &str,
     ) -> StratisResult<RenameAction<PoolUuid>> {
         validate_name(new_name)?;
-        let old_name = rename_pool_pre_idem!(self; uuid; new_name);
+        let new_name = Name::new(new_name.to_owned());
+        let old_name = rename_pool_pre_idem!(self; uuid; new_name.clone());
 
-        let (_, mut pool) = self
-            .pools
+        let mut guard = self.pools.write_all().await;
+
+        let (_, mut pool) = guard
             .remove_by_uuid(uuid)
             .expect("Must succeed since self.pools.get_by_uuid() returned a value");
 
-        let new_name = Name::new(new_name.to_owned());
-        if let Err(err) = pool.write_metadata(&new_name) {
-            self.pools.insert(old_name, uuid, pool);
+        let cloned_new_name = new_name.clone();
+        let (res, pool) = spawn_blocking!((pool.write_metadata(&cloned_new_name), pool))?;
+        if let Err(err) = res {
+            guard.insert(old_name, uuid, pool);
             Err(err)
         } else {
-            self.pools.insert(new_name, uuid, pool);
-            let (new_name, pool) = self.pools.get_by_uuid(uuid).expect("Inserted above");
+            guard.insert(new_name, uuid, pool);
+            let (new_name, pool) = guard.get_by_uuid(uuid).expect("Inserted above");
             pool.udev_pool_change(&new_name);
             Ok(RenameAction::Renamed(uuid))
         }
     }
 
-    fn unlock_pool(
-        &mut self,
+    async fn unlock_pool(
+        &self,
         pool_uuid: PoolUuid,
         unlock_method: UnlockMethod,
     ) -> StratisResult<SetUnlockAction<DevUuid>> {
-        let unlocked = self
-            .liminal_devices
-            .unlock_pool(&self.pools, pool_uuid, unlock_method)?;
+        let mut ld_guard = self.liminal_devices.write().await;
+        let pools_read_all = self.pools.read_all().await;
+        let unlocked =
+            spawn_blocking!(ld_guard.unlock_pool(&*pools_read_all, pool_uuid, unlock_method,))??;
         Ok(SetUnlockAction::new(unlocked))
     }
 
-    fn get_pool(&self, uuid: PoolUuid) -> Option<(Name, &Self::Pool)> {
-        get_pool!(self; uuid)
+    async fn get_pool(
+        &self,
+        key: LockKey<PoolUuid>,
+    ) -> Option<SomeLockReadGuard<PoolUuid, Self::Pool>> {
+        get_pool!(self; key)
     }
 
-    fn get_mut_pool(&mut self, uuid: PoolUuid) -> Option<(Name, &mut Self::Pool)> {
-        get_mut_pool!(self; uuid)
+    async fn get_mut_pool(
+        &self,
+        key: LockKey<PoolUuid>,
+    ) -> Option<SomeLockWriteGuard<PoolUuid, Self::Pool>> {
+        get_mut_pool!(self; key)
     }
 
-    fn locked_pools(&self) -> HashMap<PoolUuid, LockedPoolInfo> {
-        self.liminal_devices.locked_pools()
+    async fn locked_pools(&self) -> HashMap<PoolUuid, LockedPoolInfo> {
+        self.liminal_devices.read().await.locked_pools()
     }
 
-    fn pools(&self) -> Vec<(Name, PoolUuid, &Self::Pool)> {
-        self.pools
-            .iter()
-            .map(|(name, uuid, pool)| (name.clone(), *uuid, pool))
-            .collect()
+    async fn pools(&self) -> AllLockReadGuard<PoolUuid, Self::Pool> {
+        self.pools.read_all().await
     }
 
-    fn pools_mut(&mut self) -> Vec<(Name, PoolUuid, &mut Self::Pool)> {
-        self.pools
-            .iter_mut()
-            .map(|(name, uuid, pool)| (name.clone(), *uuid, pool))
-            .collect()
+    async fn pools_mut(&self) -> AllLockWriteGuard<PoolUuid, Self::Pool> {
+        self.pools.write_all().await
     }
 
-    fn get_events(&mut self) -> StratisResult<Vec<PoolUuid>> {
+    async fn get_events(&self) -> StratisResult<HashSet<PoolUuid>> {
         let device_list: HashMap<_, _> = get_dm()
             .list_devices()?
             .into_iter()
@@ -300,8 +483,10 @@ impl Engine for StratEngine {
             })
             .collect();
 
-        let mut changed = Vec::new();
-        for (_, pool_uuid, pool) in self.pools.iter_mut() {
+        let mut changed = HashSet::new();
+        let read_pools = self.pools.read_all().await;
+        let mut write_event_nrs = self.watched_dev_last_event_nrs.write().await;
+        for (_, pool_uuid, pool) in read_pools.iter() {
             let dev_names = pool.get_eventing_dev_names(*pool_uuid);
             let event_nrs = device_list
                 .iter()
@@ -314,141 +499,42 @@ impl Engine for StratEngine {
                 })
                 .collect::<HashMap<_, _>>();
 
-            if self.watched_dev_last_event_nrs.get(pool_uuid) != Some(&event_nrs) {
-                changed.push(*pool_uuid);
+            if write_event_nrs.get(pool_uuid) != Some(&event_nrs) {
+                changed.insert(*pool_uuid);
             }
 
-            self.watched_dev_last_event_nrs
-                .insert(*pool_uuid, event_nrs);
+            write_event_nrs.insert(*pool_uuid, event_nrs);
         }
 
         Ok(changed)
     }
 
-    fn pool_evented(
-        &mut self,
-        pools: Option<&Vec<PoolUuid>>,
-    ) -> StratisResult<HashMap<PoolUuid, ThinPoolDiff>> {
-        fn handle_eventing(
-            name: &Name,
-            uuid: PoolUuid,
-            pool: &mut StratPool,
-            changed: &mut HashMap<PoolUuid, ThinPoolDiff>,
-            errors: &mut Vec<StratisError>,
-        ) {
-            match pool.event_on(uuid, name) {
-                Ok(diff) => {
-                    if diff.is_changed() {
-                        changed.insert(uuid, diff);
-                    }
-                }
-                Err(e) => {
-                    errors.push(e);
-                }
-            }
-        }
-
-        let mut changed = HashMap::default();
-        let mut errors = Vec::new();
-
+    async fn pool_evented(
+        &self,
+        pools: Option<&HashSet<PoolUuid>>,
+    ) -> HashMap<PoolUuid, StratPoolDiff> {
         match pools {
-            Some(ps) => {
-                for uuid in ps {
-                    if let Some((name, pool)) = self.pools.get_mut_by_uuid(*uuid) {
-                        handle_eventing(&name, *uuid, pool, &mut changed, &mut errors);
-                    } else {
-                        errors.push(StratisError::Msg(format!(
-                            "Pool with UUID {} could not be found",
-                            uuid
-                        )));
-                    }
-                }
-            }
-            None => {
-                for (name, uuid, pool) in self.pools.iter_mut() {
-                    handle_eventing(name, *uuid, pool, &mut changed, &mut errors);
-                }
-            }
-        }
-
-        if errors.is_empty() {
-            Ok(changed)
-        } else {
-            let msg = if changed.is_empty() {
-                "The following errors were reported while handling devicemapper eventing"
-                    .to_string()
-            } else {
-                format!("Operations on pools with UUIDs {:?} succeeded but the following errors were also reported while handling devicemapper eventing", changed)
-            };
-            Err(StratisError::BestEffortError(msg, errors))
+            Some(ps) => self.pool_evented_dm(ps).await,
+            None => self.pool_evented_timer().await,
         }
     }
 
-    fn fs_evented(
-        &mut self,
-        pools: Option<&Vec<PoolUuid>>,
-    ) -> StratisResult<HashMap<FilesystemUuid, StratFilesystemDiff>> {
-        let mut changed = HashMap::default();
-        let mut errors = Vec::new();
-
-        fn handle_eventing(
-            name: &Name,
-            uuid: PoolUuid,
-            pool: &mut StratPool,
-            changed: &mut HashMap<FilesystemUuid, StratFilesystemDiff>,
-            errors: &mut Vec<StratisError>,
-        ) {
-            match pool.fs_event_on(uuid, name) {
-                Ok(newly_changed) => {
-                    changed.extend(newly_changed);
-                }
-                Err(e) => {
-                    errors.push(e);
-                }
-            }
-        }
-
+    async fn fs_evented(
+        &self,
+        pools: Option<&HashSet<PoolUuid>>,
+    ) -> HashMap<FilesystemUuid, StratFilesystemDiff> {
         match pools {
-            Some(ps) => {
-                for uuid in ps {
-                    if let Some((name, pool)) = self.pools.get_mut_by_uuid(*uuid) {
-                        handle_eventing(&name, *uuid, pool, &mut changed, &mut errors);
-                    } else {
-                        errors.push(StratisError::Msg(format!(
-                            "Pool with UUID {} could not be found",
-                            uuid
-                        )));
-                    }
-                }
-            }
-            None => {
-                for (name, uuid, pool) in self.pools.iter_mut() {
-                    handle_eventing(name, *uuid, pool, &mut changed, &mut errors);
-                }
-            }
-        }
-
-        if errors.is_empty() {
-            Ok(changed)
-        } else {
-            let msg = if !changed.is_empty() {
-                format!(
-                    "Operations on filesystems with UUIDs {:?} succeeded but the following errors were also reported while handling devicemapper eventing for filesystems",
-                    changed.keys().collect::<Vec<_>>(),
-                )
-            } else {
-                "The following errors were reported while handling devicemapper eventing for filesystems".to_string()
-            };
-            Err(StratisError::BestEffortError(msg, errors))
+            Some(ps) => self.fs_evented_dm(ps).await,
+            None => self.fs_evented_timer().await,
         }
     }
 
-    fn get_key_handler(&self) -> &Self::KeyActions {
-        &self.key_handler
+    async fn get_key_handler(&self) -> SharedGuard<OwnedRwLockReadGuard<Self::KeyActions>> {
+        self.key_handler.read().await
     }
 
-    fn get_key_handler_mut(&mut self) -> &mut Self::KeyActions {
-        &mut self.key_handler
+    async fn get_key_handler_mut(&self) -> ExclusiveGuard<OwnedRwLockWriteGuard<Self::KeyActions>> {
+        self.key_handler.write().await
     }
 
     fn is_sim(&self) -> bool {
@@ -458,7 +544,7 @@ impl Engine for StratEngine {
 
 #[cfg(test)]
 mod test {
-    use std::{env, error::Error};
+    use std::{env, error::Error, ffi::OsStr, iter::once, path::Path};
 
     use devicemapper::{Bytes, Sectors};
 
@@ -468,36 +554,44 @@ mod test {
             backstore::crypt_metadata_size,
             cmd,
             tests::{crypt, dm_stratis_devices_remove, loopbacked, real, FailDevice},
+            udev::{CRYPTO_FS_TYPE, FS_TYPE_KEY},
+        },
+        types::{
+            ActionAvailability, EngineAction, KeyDescription, UdevEngineDevice, UdevEngineEvent,
         },
-        types::{ActionAvailability, EngineAction, KeyDescription},
     };
 
     use super::*;
 
     /// Verify that a pool rename causes the pool metadata to get the new name.
     fn test_pool_rename(paths: &[&Path]) {
-        let mut engine = StratEngine::initialize().unwrap();
+        let engine = StratEngine::initialize().unwrap();
 
         let name1 = "name1";
-        let uuid1 = engine
-            .create_pool(name1, paths, None, None)
+        let uuid1 = test_async!(engine.create_pool(name1, paths, None))
             .unwrap()
             .changed()
             .unwrap();
 
+        let events = generate_events!();
+        test_async!(engine.handle_events(events));
+
         let fs_name1 = "testfs1";
         let fs_name2 = "testfs2";
-        let (_, pool) = engine.pools.get_mut_by_uuid(uuid1).unwrap();
-        let fs_uuid1 = pool
-            .create_filesystems(name1, uuid1, &[(fs_name1, None)])
-            .unwrap()
-            .changed()
-            .unwrap();
-        let fs_uuid2 = pool
-            .create_filesystems(name1, uuid1, &[(fs_name2, None)])
-            .unwrap()
-            .changed()
-            .unwrap();
+        let (fs_uuid1, fs_uuid2) = {
+            let mut pool = test_async!(engine.get_mut_pool(LockKey::Uuid(uuid1))).unwrap();
+            let fs_uuid1 = pool
+                .create_filesystems(name1, uuid1, &[(fs_name1, None)])
+                .unwrap()
+                .changed()
+                .unwrap();
+            let fs_uuid2 = pool
+                .create_filesystems(name1, uuid1, &[(fs_name2, None)])
+                .unwrap()
+                .changed()
+                .unwrap();
+            (fs_uuid1, fs_uuid2)
+        };
 
         cmd::udev_settle().unwrap();
 
@@ -505,7 +599,7 @@ mod test {
         assert!(Path::new(&format!("/dev/stratis/{}/{}", name1, fs_name2)).exists());
 
         let name2 = "name2";
-        let action = engine.rename_pool(uuid1, name2).unwrap();
+        let action = test_async!(engine.rename_pool(uuid1, name2)).unwrap();
 
         cmd::udev_settle().unwrap();
 
@@ -514,31 +608,37 @@ mod test {
         assert!(Path::new(&format!("/dev/stratis/{}/{}", name2, fs_name1)).exists());
         assert!(Path::new(&format!("/dev/stratis/{}/{}", name2, fs_name2)).exists());
 
-        let (_, pool) = engine.pools.get_mut_by_uuid(uuid1).unwrap();
-        pool.destroy_filesystems(
-            name2,
-            fs_uuid1
-                .into_iter()
-                .map(|(_, u, _)| u)
-                .collect::<Vec<_>>()
-                .as_slice(),
-        )
-        .unwrap();
-        pool.destroy_filesystems(
-            name2,
-            fs_uuid2
-                .into_iter()
-                .map(|(_, u, _)| u)
-                .collect::<Vec<_>>()
-                .as_slice(),
-        )
-        .unwrap();
+        {
+            let mut pool = test_async!(engine.get_mut_pool(LockKey::Uuid(uuid1))).unwrap();
+            pool.destroy_filesystems(
+                name2,
+                fs_uuid1
+                    .into_iter()
+                    .map(|(_, u, _)| u)
+                    .collect::<Vec<_>>()
+                    .as_slice(),
+            )
+            .unwrap();
+            pool.destroy_filesystems(
+                name2,
+                fs_uuid2
+                    .into_iter()
+                    .map(|(_, u, _)| u)
+                    .collect::<Vec<_>>()
+                    .as_slice(),
+            )
+            .unwrap();
+        }
 
         assert_eq!(action, RenameAction::Renamed(uuid1));
         engine.teardown().unwrap();
 
         let engine = StratEngine::initialize().unwrap();
-        let pool_name: String = engine.get_pool(uuid1).unwrap().0.to_owned();
+        let pool_name: String = test_async!(engine.get_pool(LockKey::Uuid(uuid1)))
+            .unwrap()
+            .as_tuple()
+            .0
+            .to_owned();
         assert_eq!(pool_name, name2);
     }
 
@@ -572,38 +672,39 @@ mod test {
 
         let (paths1, paths2) = paths.split_at(paths.len() / 2);
 
-        let mut engine = StratEngine::initialize().unwrap();
+        let engine = StratEngine::initialize().unwrap();
 
         let name1 = "name1";
-        let uuid1 = engine
-            .create_pool(name1, paths1, None, None)
+        let uuid1 = test_async!(engine.create_pool(name1, paths1, None))
             .unwrap()
             .changed()
             .unwrap();
 
         let name2 = "name2";
-        let uuid2 = engine
-            .create_pool(name2, paths2, None, None)
+        let uuid2 = test_async!(engine.create_pool(name2, paths2, None))
             .unwrap()
             .changed()
             .unwrap();
 
-        assert!(engine.get_pool(uuid1).is_some());
-        assert!(engine.get_pool(uuid2).is_some());
+        let events = generate_events!();
+        test_async!(engine.handle_events(events));
+
+        assert!(test_async!(engine.get_pool(LockKey::Uuid(uuid1))).is_some());
+        assert!(test_async!(engine.get_pool(LockKey::Uuid(uuid2))).is_some());
 
         engine.teardown().unwrap();
 
         let engine = StratEngine::initialize().unwrap();
 
-        assert!(engine.get_pool(uuid1).is_some());
-        assert!(engine.get_pool(uuid2).is_some());
+        assert!(test_async!(engine.get_pool(LockKey::Uuid(uuid1))).is_some());
+        assert!(test_async!(engine.get_pool(LockKey::Uuid(uuid2))).is_some());
 
         engine.teardown().unwrap();
 
         let engine = StratEngine::initialize().unwrap();
 
-        assert!(engine.get_pool(uuid1).is_some());
-        assert!(engine.get_pool(uuid2).is_some());
+        assert!(test_async!(engine.get_pool(LockKey::Uuid(uuid1))).is_some());
+        assert!(test_async!(engine.get_pool(LockKey::Uuid(uuid2))).is_some());
 
         engine.teardown().unwrap();
     }
@@ -630,7 +731,7 @@ mod test {
         F: Fn(&mut StratPool) -> Result<(), Box<dyn Error>>,
     {
         fn needs_clean_up<F>(
-            mut engine: StratEngine,
+            engine: StratEngine,
             uuid: PoolUuid,
             fail_device: &FailDevice,
             operation: F,
@@ -638,21 +739,24 @@ mod test {
         where
             F: Fn(&mut StratPool) -> Result<(), Box<dyn Error>>,
         {
-            let (_, pool) = engine
-                .get_mut_pool(uuid)
-                .ok_or_else(|| Box::new(StratisError::Msg("Pool must be present".to_string())))?;
+            {
+                let mut pool =
+                    test_async!(engine.get_mut_pool(LockKey::Uuid(uuid))).ok_or_else(|| {
+                        Box::new(StratisError::Msg("Pool must be present".to_string()))
+                    })?;
 
-            fail_device.start_failing(*Bytes(u128::from(crypt_metadata_size())).sectors())?;
-            if operation(pool).is_ok() {
-                return Err(Box::new(StratisError::Msg(
-                    "Clevis initialization should have failed".to_string(),
-                )));
-            }
+                fail_device.start_failing(*Bytes(u128::from(crypt_metadata_size())).sectors())?;
+                if operation(&mut pool).is_ok() {
+                    return Err(Box::new(StratisError::Msg(
+                        "Clevis initialization should have failed".to_string(),
+                    )));
+                }
 
-            if pool.avail_actions() != ActionAvailability::Full {
-                return Err(Box::new(StratisError::Msg(
-                    "Pool should have rolled back the change entirely".to_string(),
-                )));
+                if pool.avail_actions() != ActionAvailability::Full {
+                    return Err(Box::new(StratisError::Msg(
+                        "Pool should have rolled back the change entirely".to_string(),
+                    )));
+                }
             }
 
             fail_device.stop_failing()?;
@@ -662,23 +766,58 @@ mod test {
             Ok(())
         }
 
-        let mut engine = StratEngine::initialize()?;
-        let uuid = engine
-            .create_pool(name, paths_with_fail_device, None, Some(encryption_info))?
-            .changed()
-            .ok_or_else(|| {
-                Box::new(StratisError::Msg(
-                    "Pool should be newly created".to_string(),
-                ))
-            })?;
+        let engine = StratEngine::initialize()?;
+        let uuid =
+            test_async!(engine.create_pool(name, paths_with_fail_device, Some(encryption_info)))?
+                .changed()
+                .ok_or_else(|| {
+                    Box::new(StratisError::Msg(
+                        "Pool should be newly created".to_string(),
+                    ))
+                })?;
+
+        let mut events = generate_events!();
+
+        // Hack because fail device does not show up as LUKS2 device in udev
+        let cxt = libudev::Context::new().expect("Creating udev context should succeed");
+        let mut enumerator =
+            libudev::Enumerator::new(&cxt).expect("Creating enumerator should succeed");
+        enumerator.match_is_initialized().unwrap();
+        let mut devices = enumerator
+            .scan_devices()
+            .expect("Scanning udev devices should succeed");
+
+        let fail_udev = devices
+            .find(|dev| dev.devnode() == Some(&fail_device.as_path().canonicalize().unwrap()))
+            .expect("Fail device must be in udev database");
+
+        events.push(UdevEngineEvent::new(
+            libudev::EventType::Add,
+            UdevEngineDevice::new(
+                fail_udev.is_initialized(),
+                fail_udev.devnode().map(|p| p.to_owned()),
+                fail_udev.devnum(),
+                fail_udev
+                    .properties()
+                    .map(|prop| (Box::from(prop.name()), Box::from(prop.value())))
+                    .chain(once((
+                        Box::from(OsStr::new(FS_TYPE_KEY)),
+                        Box::from(OsStr::new(CRYPTO_FS_TYPE)),
+                    )))
+                    .collect::<HashMap<_, _>>(),
+            ),
+        ));
+
+        test_async!(engine.handle_events(events));
 
         let res = needs_clean_up(engine, uuid, fail_device, operation);
+
         dm_stratis_devices_remove()?;
         res?;
 
-        let mut engine = StratEngine::initialize()?;
-        engine.unlock_pool(uuid, unlock_method)?;
-        engine.destroy_pool(uuid)?;
+        let engine = StratEngine::initialize()?;
+        test_async!(engine.unlock_pool(uuid, unlock_method))?;
+        test_async!(engine.destroy_pool(uuid))?;
         engine.teardown()?;
 
         Ok(())
